import numpy as np
import scipy as sp

from ..constants import T_CMB, k_B
from ..functions.radiometry import (
    inverse_planck_spectrum,
    inverse_rayleigh_jeans_spectrum,
    planck_spectrum,
    rayleigh_jeans_spectrum,
)  # noqa


def identity(x: float, **kwargs):
    return x


def rayleigh_jeans_temperature_to_brightness_temperature(T_RJ, nu, **kwargs):
    I_nu = rayleigh_jeans_spectrum(T_RJ=T_RJ, nu=nu)
    return inverse_planck_spectrum(I_nu=I_nu, nu=nu)


def brightness_temperature_to_rayleigh_jeans_temperature(T_b, nu, **kwargs):
    I_nu = planck_spectrum(T_b=T_b, nu=nu)
    return inverse_rayleigh_jeans_spectrum(I_nu=I_nu, nu=nu)


def rayleigh_jeans_temperature_to_radiant_flux(
    T_RJ,
    band,
    spectrum=None,
    **kwargs,
):
    """
    nu: frequency, in Hz
    band: response to a Rayleigh-Jeans source
    """
    if spectrum:
        integral = band.compute_nu_integral(
            spectrum=spectrum,
            zenith_pwv=kwargs["zenith_pwv"],
            base_temperature=kwargs["base_temperature"],
            elevation=kwargs["elevation"],
        )

    else:
        integral = band.compute_nu_integral(spectrum=None)

    return T_RJ * k_B * integral


def radiant_flux_to_rayleigh_jeans_temperature(P, band, spectrum=None, **kwargs):
    """
    nu: frequency, in Hz
    passband: response to a Rayleigh-Jeans source
    """
    if spectrum:
        integral = band.compute_nu_integral(
            spectrum=spectrum,
            zenith_pwv=kwargs["zenith_pwv"],
            base_temperature=kwargs["base_temperature"],
            elevation=kwargs["elevation"],
        )

    else:
        integral = band.compute_nu_integral(spectrum=None)

    return P / (k_B * integral)


def brightness_temperature_to_radiant_flux(T_b, band, polarized=False, spectrum=None, **kwargs):
    # test_T_b = np.linspace(np.min(T_b) - 1e-6, np.max(T_b) + 1e-6, 2)

    stokes_factor = 0.5 if polarized else 1.0

    if spectrum:
        T_RJ = inverse_rayleigh_jeans_spectrum(
            planck_spectrum(T_b=np.atleast_2d(T_b), nu=spectrum.side_nu[:, None]),
            nu=spectrum.side_nu[:, None],
        )

        integral = np.trapezoid(
            y=T_RJ[None, None, None] * (np.exp(-spectrum._opacity) * band.passband(spectrum.side_nu))[..., None],
            x=spectrum.side_nu,
            axis=-2,
        )
        points = spectrum.points[:3]
        xi = (
            kwargs["base_temperature"],
            kwargs["zenith_pwv"],
            kwargs["elevation"],
        )
        integral = sp.interpolate.RegularGridInterpolator(points, integral)(xi)

    else:
        T_RJ = inverse_rayleigh_jeans_spectrum(
            planck_spectrum(T_b=np.atleast_2d(T_b), nu=band.nu[:, None]),
            nu=band.nu[:, None],
        )
        integral = np.trapezoid(y=T_RJ * band.passband(band.nu[:, None]), x=band.nu, axis=-2)

    return k_B * stokes_factor * integral


def radiant_flux_to_brightness_temperature(P, **kwargs):
    raise NotImplementedError()


# def dP_dT_CMB(band, spectrum, eps=1e-3, **kwargs):
#     test_T_b = np.array([T_CMB - eps / 2, T_CMB + eps / 2])
#     test_P = brightness_temperature_to_radiant_flux(test_T_b, band=band, spectrum=spectrum, **kwargs)
#     return (test_P[..., 1] - test_P[..., 0]) / eps


def dP_dT_CMB(band, spectrum, eps=1e-6, **kwargs):
    P_lo = brightness_temperature_to_radiant_flux(T_CMB - eps / 2, band=band, spectrum=spectrum, **kwargs)
    P_hi = brightness_temperature_to_radiant_flux(T_CMB + eps / 2, band=band, spectrum=spectrum, **kwargs)

    return (P_hi - P_lo) / eps


def cmb_temperature_anisotropy_to_radiant_flux(T_b, band, spectrum=None, **kwargs):
    return T_b * dP_dT_CMB(band=band, spectrum=spectrum, **kwargs)


def radiant_flux_to_cmb_temperature_anisotropy(P, band, spectrum=None, **kwargs):
    return P / dP_dT_CMB(band=band, spectrum=spectrum, **kwargs)


def rayleigh_jeans_temperature_to_spectral_flux_density_per_pixel(T_RJ: float, nu: float, pixel_area: float, **kwargs):
    """
    T_RJ: Rayleigh-Jeans temperature, in Kelvin
    nu: frequency, in Hz
    res: resolution, in radians
    """
    return 1e26 * rayleigh_jeans_spectrum(T_RJ=T_RJ, nu=nu) * pixel_area


def spectral_flux_density_per_pixel_to_rayleigh_jeans_temperature(E: float, nu: float, pixel_area: float, **kwargs):
    """
    T_RJ: Rayleigh-Jeans temperature, in Jy/pixel
    nu: frequency, in Hz
    res: resolution, in radians
    """
    I_nu = 1e-26 * E / pixel_area
    return inverse_rayleigh_jeans_spectrum(I_nu=I_nu, nu=nu)


def cmb_temperature_anisotropy_to_radiant_flux_slope(
    band,
    eps: float = 1e-3,
    **kwargs,
):
    test_T_b = T_CMB + np.array([[-eps / 2], [+eps / 2]])
    T_RJ = inverse_rayleigh_jeans_spectrum(
        planck_spectrum(T_b=test_T_b, nu=band.nu),
        nu=band.nu,
    )
    P = k_B * np.trapezoid(T_RJ * band.passband(band.nu), x=band.nu)
    return (P[1] - P[0]) / eps


def cmb_temperature_anisotropy_to_rayleigh_jeans_temperature(
    delta_T: float,
    band,
    **kwargs,
):
    dP_dTCMB = cmb_temperature_anisotropy_to_radiant_flux_slope(band=band)
    return radiant_flux_to_rayleigh_jeans_temperature(
        dP_dTCMB * delta_T,
        band=band,
    )


def rayleigh_jeans_temperature_to_cmb_temperature_anisotropy(
    T_RJ: float,
    band,
    **kwargs,
):
    dP_dTCMB = cmb_temperature_anisotropy_to_radiant_flux_slope(band=band)
    return rayleigh_jeans_temperature_to_radiant_flux(T_RJ, band=band) / dP_dTCMB


def spectral_radiance_to_spectral_flux_density_per_pixel(E: float, pixel_area: float, **kwargs):
    """
    E: Spectral flux density per area, in Jy/sr
    pixel_area: pixel area, in steradians
    """
    return E * pixel_area


def spectral_flux_density_per_pixel_to_spectral_radiance(E: float, pixel_area: float, **kwargs):
    """
    E: Spectral flux density per pixel, in Jy/pixel
    pixel_area: pixel area, in steradians
    """
    return E / pixel_area


def spectral_flux_density_per_beam_to_spectral_flux_density_per_pixel(
    E: float, pixel_area: float, beam_area: float, **kwargs
):
    """
    E: Spectral flux density per area, in Jy/sr
    pixel_area: pixel area, in steradians
    """
    return E * pixel_area / beam_area


def spectral_flux_density_per_pixel_to_spectral_flux_density_per_beam(
    E: float, pixel_area: float, beam_area: float, **kwargs
):
    """
    E: Spectral flux density per pixel, in Jy/pixel
    pixel_area: pixel area, in steradians
    """
    return E * beam_area / pixel_area



import os
import re

import numpy as np
import pandas as pd
import yaml

from .prefixes import PREFIXES

here, this_filename = os.path.split(__file__)

with open(f"{here}/quantities.yml") as f:
    QUANTITIES = yaml.safe_load(f)

units_entries = {}
for q, q_config in QUANTITIES.items():
    for unit, unit_entry in q_config.pop("units").items():
        units_entries[unit] = {**unit_entry, "quantity": q, "quantity_default_unit": q_config["default_unit"]}
        units_entries[unit]["aliases"] = [unit, *units_entries[unit].get("aliases", [])]

UNITS = pd.DataFrame(units_entries).fillna("").T
UNITS["factor"] = UNITS["factor"].astype(float)

prefixes_phrase = r"|".join(PREFIXES.index)
base_units_phrase = r"|".join([alias for _, entry in UNITS.iterrows() for alias in entry.aliases]).replace("^", "\\^")
units_pattern = re.compile(rf"^(?P<prefix>({prefixes_phrase}))(?P<base_unit>{base_units_phrase})$")  # noqa


def parse_units(u):
    match = units_pattern.search(u)
    if match is None:
        raise ValueError(
            f"Invalid units '{u}'. Valid units are a combination of an SI prefix "
            f"(one of {prefixes_phrase}) and a base unit (one of {base_units_phrase}).",
        )

    units_dict = match.groupdict()
    prefix = PREFIXES.loc[units_dict["prefix"]]
    for _, entry in UNITS.iterrows():
        if units_dict["base_unit"] in entry["aliases"]:
            base_unit = entry
            break

    units_dict["units"] = f"{prefix.name}{base_unit.name}"
    units_dict.update(base_unit.to_dict())
    units_dict["long_name"] = f"{prefix.long_name}{base_unit.long_name}"
    units_dict["math_name"] = f"{prefix.symbol_latex}{base_unit.math_name}"
    units_dict["factor"] = base_unit.factor * prefix.factor
    if not units_dict["symbol"]:
        units_dict.pop("symbol")
    return units_dict


class Quantity:
    """
    A number (or numbers) with units representing a quantity with some dimensions.
    """

    def __init__(self, value, units):
        u = parse_units(units)
        x = np.array(value) * u["factor"]  # the

        # u = parse_units(u["default_unit"]) # the

        self.q = QUANTITIES[u["quantity"]]
        natural_units = UNITS.loc[(UNITS.quantity == u["quantity"]) & (UNITS.natural)].sort_values("factor")

        abs_x = np.abs(x)
        abs_fin_x = np.where((abs_x > 0) & np.isfinite(abs_x), abs_x, np.nan)

        if (abs_fin_x > 0).any():
            fid_x = 2 * np.nanquantile(abs_fin_x, q=0.99)
            unit_index = np.digitize(fid_x, [0, *natural_units.factor.values[1:], np.inf]) - 1
            unit = natural_units.iloc[unit_index]

            power = 3 * (np.log10(fid_x / unit.factor) // 3)
            power = np.clip(power, a_min=self.min_prefix_power, a_max=self.max_prefix_power)
            prefix = PREFIXES.set_index("power").loc[power].symbol
            self.u = parse_units(f"{prefix}{unit.name}")

        else:
            self.u = parse_units(u["quantity_default_unit"])

        self.value = x / self.u["factor"]

    @property
    def units(self) -> str:
        return self.u["units"]

    # @property
    # def human_units_math(self) -> str:
    #     return parse_units(self.human_units)["math_name"]

    @property
    def min_prefix_power(self) -> bool:
        return self.q.get("min_prefix_power", -30)

    @property
    def max_prefix_power(self) -> bool:
        return self.q.get("max_prefix_power", 30)

    def __repr__(self) -> str:
        u = self.u
        if (u["prefix"] == "") and ("symbol" in u):
            units_string = u["symbol"]
        else:
            units_string = f" {self.units}"
        value_repr = f"{self.value:.04g}" if np.isscalar(self.value) else self.value
        return f"{value_repr}{units_string}"

    def to(self, units) -> float:
        u = parse_units(units)
        if u["quantity"] != self.u["quantity"]:
            raise ValueError(f"Cannot convert quantity '{self.u['quantity']}' to quantity '{u['quantity']}'.")
        return self.value * self.u["factor"] / u["factor"]

    def __getattr__(self, attr):
        try:
            return self.to(attr)
        except Exception:
            pass
        raise AttributeError(f"'Quantity' object has no attribute '{attr}'")

    def __getitem__(self, key):
        return Quantity(self.value[key], units=self.units)

    def mean(self):
        return Quantity(np.mean(self.value), units=self.u["units"])

    def __add__(self, other):
        return Quantity(self.value + getattr(other, self.units), units=self.units)

    def __sub__(self, other):
        return Quantity(self.value - getattr(other, self.units), units=self.units)

    def __format__(self, *args, **kwargs):
        return repr(self).__format__(*args, **kwargs)



# speed of light (m s^-1)
c = 2.99792458e8

# little g (m s^-2)
g = 9.806651

# Planck's constant (kg m^2 s)
h = 6.62607015e-34

# Reduced Planck's constant (kg m^2 s)
hbar = h / 6.283185307179586

# Boltzmann's constant (kg m^2 s^-2 K^-1)
k_B = 1.380649e-23

# CMB temperature (K)
T_CMB = 2.72548

MARIA_MIN_NU = 1e6  # 1 MHz
MARIA_MAX_NU = 15e12  # 150 THz



import arrow


def get_utc_day_hour(t):
    a = arrow.get(t).to("utc")
    return a.hour + a.minute / 60 + a.second / 3600


def get_utc_year_day(t):
    return int(arrow.get(t).to("utc").format("DDD")) + get_utc_day_hour(t) / 24 - 1


def get_utc_year(t):
    return arrow.get(t).to("utc").year



from __future__ import annotations

import logging
import os
import time as ttime

import arrow
import numpy as np
from tqdm import tqdm

from ..atmosphere import DEFAULT_ATMOSPHERE_KWARGS, Atmosphere
from ..cmb import CMB, DEFAULT_CMB_KWARGS, generate_cmb, get_cmb
from ..errors import PointingError
from ..instrument import Instrument
from ..io import fetch, humanize_time
from ..map import Map, load
from ..plan import Plan
from ..site import Site
from .atmosphere import AtmosphereMixin
from .base import BaseSimulation
from .cmb import CMBMixin
from .map import MapMixin
from .noise import NoiseMixin

here, this_filename = os.path.split(__file__)
logger = logging.getLogger("maria")

MIN_ELEVATION_WARN = 10  # degrees
MIN_ELEVATION_ERROR = 5  # degrees


class Simulation(BaseSimulation, AtmosphereMixin, CMBMixin, MapMixin, NoiseMixin):
    """
    A simulation of a telescope. This is what users should touch, primarily.
    """

    @classmethod
    def from_config(cls, config: dict = {}, **params):
        return cls(**params)

    def __init__(
        self,
        instrument: tuple[Instrument, str] = "test/1deg",
        plan: tuple[Plan, str] = "one_minute_zenith_stare",
        site: tuple[Site, str] = "hoagie_haven",
        atmosphere: tuple[Atmosphere, str] = None,
        cmb: tuple[CMB, str] = None,
        map: tuple[Map, str] = None,
        noise: bool = True,
        atmosphere_kwargs: dict = {},
        cmb_kwargs: dict = {},
        map_kwargs: dict = {},
        noise_kwargs: dict = {},
        progress_bars: bool = True,
        keep_mean_signal: bool = False,
    ):
        super().__init__(
            instrument=instrument,
            plan=plan,
            site=site,
            progress_bars=progress_bars,
            keep_mean_signal=keep_mean_signal,
        )

        sim_start_s = ttime.monotonic()

        self.noise = noise

        self.map_kwargs = map_kwargs
        self.noise_kwargs = noise_kwargs

        self.start = arrow.get(self.boresight.t.min()).to("utc")
        self.end = arrow.get(self.boresight.t.max()).to("utc")

        if atmosphere:
            self.atmosphere_kwargs = DEFAULT_ATMOSPHERE_KWARGS.copy()
            self.atmosphere_kwargs.update(atmosphere_kwargs)

            # do some checks
            el_min = np.atleast_1d(self.coords.el).min()
            if el_min < np.radians(MIN_ELEVATION_WARN):
                logger.warning(
                    f"Some ArrayList come within {MIN_ELEVATION_WARN} degrees of the horizon"
                    f"(el_min = {np.degrees(el_min):.01f}°)",
                )
            if el_min <= np.radians(MIN_ELEVATION_ERROR):
                raise PointingError(
                    f"Some ArrayList come within {MIN_ELEVATION_ERROR} degrees of the horizon"
                    f"(el_min = {np.degrees(el_min):.01f}°)",
                )

            self.weather_kwargs = self.atmosphere_kwargs.pop("weather") if "weather" in self.atmosphere_kwargs else {}

            self.atmosphere = Atmosphere(
                model=atmosphere,
                timestamp=self.plan.time.mean(),
                region=self.site.region,
                altitude=self.site.altitude,
                weather_kwargs=self.weather_kwargs,
                disable_progress_bars=self.disable_progress_bars,
                **self.atmosphere_kwargs,
            )

            # give it the simulation, so that it knows about pointing, site, etc. (kind of cursed)
            self.atmosphere.initialize(self)

            logger.debug(f"Initialized atmosphere simulation in {humanize_time(ttime.monotonic() - sim_start_s)}.")

        cmb_start_s = ttime.monotonic()

        if cmb:
            self.cmb_kwargs = DEFAULT_CMB_KWARGS.copy()
            self.cmb_kwargs.update(cmb_kwargs)

            if cmb in ["spectrum", "power_spectrum", "generate", "generated"]:
                for _ in tqdm(
                    range(1),
                    desc=f"Generating CMB (nside={self.cmb_kwargs['nside']})",
                    disable=self.disable_progress_bars,
                ):
                    self.cmb = generate_cmb(**self.cmb_kwargs)
            elif cmb in ["real", "planck"]:
                self.cmb = get_cmb(**self.cmb_kwargs)
            else:
                raise ValueError(f"Invalid value for cmb: '{cmb}'.")

            logger.debug(f"Initialized CMB simulation in {humanize_time(ttime.monotonic() - cmb_start_s)}.")

        map_start_s = ttime.monotonic()
        if map:
            if isinstance(map, str):
                map = load(fetch(map), **map_kwargs)
            elif not isinstance(map, Map):
                raise ValueError("")

            # the map can be frequency-naive if it is already in K_RJ
            self.map = map.to(units="K_RJ")

            if "stokes" not in self.map.dims:
                self.map = self.map.unsqueeze("stokes")

            if "nu" not in self.map.dims:
                self.map = self.map.unsqueeze("nu")

            if "t" in self.map.dims:
                if self.map.dims["t"] > 1:
                    map_start = arrow.get(self.map.t.min()).to("utc")
                    map_end = arrow.get(self.map.t.max()).to("utc")
                    if map_start > self.start:
                        logger.warning(
                            f"Beginning of map ({map_start.isoformat()[:26]}) is after the "
                            f"beginning of the simulation ({self.start.isoformat()[:26]}).",
                        )
                    if map_end < self.end:
                        logger.warning(
                            f"End of map ({map_end.isoformat()[:26]}) is before the "
                            f"end of the simulation ({self.end.isoformat()[:26]}).",
                        )
                else:
                    self.map = self.map.squeeze("t")

        logger.debug(f"Initialized map simulation in {humanize_time(ttime.monotonic() - map_start_s)}.")

        noise_start_s = ttime.monotonic()
        if noise:
            pass
        logger.debug(f"Initialized noise simulation in {humanize_time(ttime.monotonic() - noise_start_s)}.")

        logger.debug(f"Initialized simulation in {humanize_time(ttime.monotonic() - sim_start_s)}.")

    def _run(self):
        if hasattr(self, "atmosphere"):
            atmosphere_sim_start_s = ttime.monotonic()
            self._simulate_atmosphere()
            self._compute_atmospheric_emission()
            logger.debug(f"Ran atmosphere simulation in {humanize_time(ttime.monotonic() - atmosphere_sim_start_s)}.")

        if hasattr(self, "cmb"):
            cmb_sim_start_s = ttime.monotonic()
            self._simulate_cmb_emission()
            logger.debug(f"Ran CMB simulation in {humanize_time(ttime.monotonic() - cmb_sim_start_s)}.")

        if hasattr(self, "map"):
            map_sim_start_s = ttime.monotonic()
            self._sample_maps()
            logger.debug(f"Ran map simulation in {humanize_time(ttime.monotonic() - map_sim_start_s)}.")

        # number of bands are lost here
        if self.noise:
            noise_sim_start_s = ttime.monotonic()
            self._simulate_noise()
            logger.debug(f"Ran noise simulation in {humanize_time(ttime.monotonic() - noise_sim_start_s)}.")

        gain_error = np.exp(
            self.instrument.dets.gain_error * np.random.standard_normal(size=self.instrument.dets.n),
        )

        for field in self.loading:
            if field in ["noise"]:
                continue

            self.loading[field] *= gain_error[:, None]

    def __repr__(self):
        instrument_tree = "├ " + self.instrument.__repr__().replace("\n", "\n│ ")
        site_tree = "├ " + self.site.__repr__().replace("\n", "\n│ ")
        plan_tree = "├ " + self.plan.__repr__().replace("\n", "\n│ ")

        trees = []
        attrs = [attr for attr in ["atmosphere", "cmb", "map"] if getattr(self, attr, None)]
        for attr in attrs[:-1]:
            trees.append("├ " + getattr(self, attr).__repr__().replace("\n", "\n│ "))
        trees.append("└ " + getattr(self, attrs[-1]).__repr__().replace("\n", "\n  "))

        trees_string = "\n".join(trees)

        return f"""Simulation
{instrument_tree}
{site_tree}
{plan_tree}
{trees_string}"""

    @property
    def total_loading(self):
        return sum([d for d in self.loading.values()])



import copy
import logging
import os
from typing import Callable, Iterable

import astropy as ap
import dask.array as da
import h5py
import jax
import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import scipy as sp
from astropy.io import fits
from astropy.wcs import WCS
from skimage.measure import block_reduce

from ..coords import Coordinates, frames
from ..units import Quantity, parse_units
from ..utils import compute_pointing_matrix, repr_phi_theta, unpack_implicit_slice
from .base import Map

here, this_filename = os.path.split(__file__)
logger = logging.getLogger("maria")


class ProjectedMap(Map):
    """
    A rectangular map projected on the sphere. It has shape (stokes, nu, t, y, x).
    """

    def __init__(
        self,
        data: float,
        weight: float = None,
        stokes: float = None,
        nu: Iterable[float] = None,
        t: Iterable[float] = None,
        z: Iterable[float] = None,
        width: float = None,
        height: float = None,
        resolution: float = None,
        x_res: float = None,
        y_res: float = None,
        center: tuple[float, float] = (0.0, 0.0),
        beam: tuple[float, float, float] = 0.0,
        frame: str = "ra_dec",
        degrees: bool = True,
        units: str = "K_RJ",
        dtype: type = np.float32,
    ):
        # give it five dimensions

        if data.ndim < 2:
            raise ValueError("'data' must be at least two-dimensional")

        if weight is not None:
            if weight.shape != data.shape:
                raise ValueError(f"'data' {data.shape} and 'weight' {weight.shape} should have the same shape.")
        else:
            weight = da.ones_like(data)

        map_dims = {"y": data.shape[-2], "x": data.shape[-1]}

        super().__init__(
            data=data,
            weight=weight,
            stokes=stokes,
            nu=nu,
            t=t,
            z=z,
            beam=beam,
            map_dims=map_dims,
            units=units,
            degrees=degrees,
            dtype=dtype,
        )

        self.center = Quantity(tuple(center), ("deg" if degrees else "rad")).rad
        self.frame = frame

        if all(x is None for x in [width, height, resolution, x_res, y_res]):
            raise ValueError("You must pass at least one of 'width', 'height', 'resolution', 'x_res', or 'y_res'.")

        if x_res is not None:
            y_res = y_res or x_res
        elif y_res is not None:
            x_res = x_res or y_res
        elif resolution is not None:
            if not resolution > 0:
                raise ValueError("'resolution' must be positive.")
            x_res = y_res = resolution
        else:
            if width is not None:
                if not width > 0:
                    raise ValueError("'width' must be positive.")
                x_res = width / self.n_x
                if height is not None:
                    y_res = height / self.n_y
                else:
                    y_res = x_res
            else:
                # here height must not be None
                x_res = y_res = height / self.n_y

        self.x_res = np.radians(x_res) if degrees else x_res
        self.y_res = np.radians(y_res) if degrees else y_res

        if self.x_res < 0:
            self.data, self.weight = self.data[::-1], self.weight[::-1]
            self.x_res *= -1
        if self.y_res < 0:
            self.data, self.weight = self.data[::-1, :], self.weight[::-1, :]
            self.y_res *= -1

        if not hasattr(beam, "__len__"):
            beam = beam or self.resolution
            beam = (beam, beam, 0)

        if len(beam) != 3:
            raise ValueError("'beam' must be either a number or a tuple of (major, minor, angle)")

        self.beam = tuple(np.radians(beam) if degrees else beam)

        if self.u["base_unit"] == "Jy/beam":
            if not self.beam_area > 0:
                raise ValueError(
                    f"Map is given in units {self.units}, but specified beam(major, minor, angle) = {beam} has zero area"
                )

    def pointing_matrix(self, coords: Coordinates):
        offsets = coords.offsets(center=self.center, frame=self.frame)
        if "t" in self.dims:
            return compute_pointing_matrix(
                points=(offsets[..., 0], offsets[..., 1], coords._t), bins=(self.x_bins, self.y_bins[::-1], self.t_bins)
            )
        else:
            return compute_pointing_matrix(points=(offsets[..., 0], offsets[..., 1]), bins=(self.x_bins, self.y_bins[::-1]))

    @property
    def header(self):
        header = ap.io.fits.header.Header()

        header["SIMPLE"] = "T / conforms to FITS standard"
        header["BITPIX"] = "-32 / array data type"
        header["NAXIS"] = len(self.dims)
        for dim_index, n in enumerate(list(self.dims.values())):
            header[f"NAXIS{dim_index + 1}"] = n

        if "nu" in self.dims:
            header["RESTFREQ"] = self.nu.mean()

        header["CDELT1"] = -1 * np.degrees(self.x_res)  # degrees
        header["CDELT2"] = np.degrees(self.y_res)  # degrees
        header["CRPIX1"] = self.data.shape[-1] // 2
        header["CRPIX2"] = self.data.shape[-2] // 2
        header["BUNITS"] = self.units

        # specify x center
        CTYPE1 = self.frame_data["FITS_phi"]
        header["CTYPE1"] = f"{CTYPE1}{(5 - len(CTYPE1)) * '-'}SIN"
        header["CRVAL1"] = np.degrees(self.center[0])
        header["CUNIT1"] = "deg     "

        # center y center
        CTYPE2 = self.frame_data["FITS_theta"]
        header["CTYPE2"] = f"{CTYPE2}{(5 - len(CTYPE2)) * '-'}SIN"
        header["CRVAL2"] = np.degrees(self.center[1])
        header["CUNIT2"] = "deg     "

        return header

    def __getattr__(self, attr):
        broadcasted_attrs = ["STOKES", "NU", "T", "Y", "X"]
        if attr in broadcasted_attrs:
            broadcasted_attr_values = np.meshgrid(self.stokes, self.nu, self.t, self.y_side, self.x_side)
            return broadcasted_attr_values[broadcasted_attrs.index(attr)]

        raise AttributeError(f"'ProjectedMap' object has no attribute '{attr}'")

    def __getitem__(self, key):
        if not hasattr(key, "__iter__"):
            key = (key,)

        explicit_slices = unpack_implicit_slice(key, ndims=self.ndim)

        package = self.to("K_RJ").package()
        package["data"] = package["data"][key]
        package["weight"] = package["weight"][key]

        for axis, (dim, naxis) in enumerate(self.dims.items()):
            if dim in "xy":
                continue
            package[dim] = package[dim][explicit_slices[axis]]

        package["y_res"] *= explicit_slices[-2].step or 1
        package["x_res"] *= explicit_slices[-1].step or 1

        return ProjectedMap(**package).to(self.units)

    def downsample(self, reduce: tuple, func: Callable = np.mean):
        if reduce:
            if len(reduce) != 4:
                raise ValueError("'reduce' must be a four-tuple of ints (nu, t, y, x)")

            package = self.to("K_RJ").package()
            package["data"] = block_reduce(package["data"].compute(), block_size=(1, *reduce), func=func)
            package["weight"] = block_reduce(package["weight"].compute(), block_size=(1, *reduce), func=func)

            *_, new_n_y, new_n_x = package["data"].shape

            package["nu"] = block_reduce(package["nu"], block_size=reduce[0])
            package["t"] = block_reduce(package["t"], block_size=reduce[1])
            package["x_res"] *= self.n_x / new_n_x
            package["y_res"] *= self.n_y / new_n_y

            return ProjectedMap(**package).to(self.units)

    @property
    def frame_data(self):
        return frames[self.frame]

    @property
    def points(self):
        return np.stack(np.meshgrid(self.y_side, self.x_side, indexing="ij"), axis=-1)

    def __repr__(self):
        cphi_repr, ctheta_repr = repr_phi_theta(*self.center, frame=self.frame)
        return f"""{self.__class__.__name__}:
  shape{self.dims_string}: {self.data.shape}
  stokes: {self.stokes if "stokes" in self.dims else "naive"}
  nu: {Quantity(self.nu, "Hz") if "nu" in self.dims else "naive"}
  t: {Quantity(self.t, "s") if "t" in self.dims else "naive"}
  z: {self.z if "z" in self.dims else "naive"}
  quantity: {self.u["quantity"]}
  units: {self.units}
    min: {np.nanmin(self.data).compute():.03e}
    max: {np.nanmax(self.data).compute():.03e}
  center:
    {cphi_repr}
    {ctheta_repr}
  size(y, x): ({Quantity(self.height, "rad")}, {Quantity(self.width, "rad")})
  resolution(y, x): ({Quantity(self.y_res, "rad")}, {Quantity(self.x_res, "rad")})
  beam(maj, min, rot): ({Quantity(self.beam[0], "rad")}, {Quantity(self.beam[1], "rad")}, {Quantity(self.beam[2], "rad")})
  memory: {Quantity(self.data.nbytes + self.weight.nbytes, "B")}"""

    def package(self):
        package = copy.deepcopy(
            {
                "degrees": True,
                "data": self.data,
                "weight": self.weight,
                "stokes": self.stokes,
                "nu": self.nu,
                "t": self.t,
                "center": np.degrees(self.center),
                "frame": self.frame,
                "units": self.units,
                "x_res": np.degrees(self.x_res),
                "y_res": np.degrees(self.y_res),
                "beam": np.degrees(self.beam),
            }
        )

        for dim in ["stokes", "nu", "t"]:
            if dim not in self.dims:
                package.pop(dim)

        return package

    @property
    def resolution(self):
        if not np.isclose(self.x_res, self.y_res, rtol=1e-3):
            RuntimeError(
                "Cannot return attribute 'resolution'; ProjectedMap has x-resolution"
                f" {np.degrees(self.x_res):.02f}° and y-resolution {np.degrees(self.x_res):.02f}°."
            )
        return self.x_res

    @property
    def width(self):
        return self.x_res * self.n_x

    @property
    def height(self):
        return self.y_res * self.n_y

    @property
    def n_y(self):
        return self.data.shape[-2]

    @property
    def n_x(self):
        return self.data.shape[-1]

    @property
    def x_bins(self):
        return self.width * np.linspace(-0.5, 0.5, self.n_x + 1)

    @property
    def y_bins(self):
        """
        The negative is so that
        """
        return -self.height * np.linspace(-0.5, 0.5, self.n_y + 1)

    @property
    def x_side(self):
        return (self.x_bins[:-1] + self.x_bins[1:]) / 2

    @property
    def y_side(self):
        return (self.y_bins[:-1] + self.y_bins[1:]) / 2

    def smooth(self, sigma: float = None, fwhm: float = None):
        if not (sigma is None) ^ (fwhm is None):
            raise ValueError("You must supply exactly one of 'sigma' or 'fwhm'.")

        package = self.package()

        sigma = sigma if sigma is not None else fwhm / np.sqrt(8 * np.log(2))
        x_sigma_pixels = sigma / self.x_res
        y_sigma_pixels = sigma / self.y_res
        package["data"] = sp.ndimage.gaussian_filter(self.data, sigma=(y_sigma_pixels, x_sigma_pixels), axes=(-2, -1))

        return type(self)(**package)

    def plot_slice(
        self,
        fig,
        ax,
        nu_index: int = 0,
        t_index: int = 0,
        stokes: str = "I",
        cmap: str = "default",
        rel_vmin: float = 0.005,
        rel_vmax: float = 0.995,
    ):
        d = self.data
        w = self.weight

        logger.debug(f"Plotting map slice (stokes={stokes}, nu_index={nu_index}, t_index={t_index})")

        if "stokes" in self.dims:
            if stokes not in self.stokes:
                raise ValueError(
                    f"Invalid stokes parameter '{stokes}'; available Stokes parameters for this map are {self.stokes}."
                )
            stokes_index = self.stokes.index(stokes)
            d, w = d[stokes_index], w[stokes_index]

        if "nu" in self.dims:
            d, w = d[nu_index], w[nu_index]

        if "t" in self.dims:
            d, w = d[t_index], w[t_index]

        if cmap == "default":
            cmap = "CMRmap" if stokes == "I" else "cmb"

        map_qdata = Quantity(d.compute(), units=self.units)
        subset = np.random.choice(d.size, size=min(d.size, 20000), replace=False)
        vmin, vmax = np.nanquantile(
            map_qdata.value.ravel()[subset],
            weights=w.ravel()[subset].compute(),
            q=[rel_vmin, rel_vmax],
            method="inverted_cdf",
        )

        # print(vmin, vmax)

        # assert False

        X = np.r_[self.x_bins, self.y_bins]
        grid_u = Quantity(X, "rad").u
        grid_center = Quantity(self.center, "rad")

        x = Quantity(self.x_bins, "rad")
        y = Quantity(self.y_bins, "rad")

        # fig = plt.figure(figsize=(6, 6), dpi=256, constrained_layout=True)
        # ax = fig.add_subplot(nx, ny, index, projection=WCS(header))

        ax.pcolormesh(
            getattr(x, grid_u["units"]),
            getattr(y, grid_u["units"]),
            map_qdata.value,
            cmap=cmap,
            vmin=vmin,
            vmax=vmax,
        )

        ax.grid(color="white", ls="solid", lw=5e-1)

        dummy_map = mpl.cm.ScalarMappable(
            mpl.colors.Normalize(vmin=vmin, vmax=vmax),
            cmap=cmap,
        )

        cbar = fig.colorbar(
            dummy_map,
            ax=ax,
            shrink=0.75,
            aspect=16,
            location="right",
            pad=0,
        )

        label = rf"${map_qdata.u['math_name']}$"
        if "stokes" in self.dims:
            label += f" Stokes {stokes}"
        if "nu" in self.dims:
            label += f" at {Quantity(self.nu[nu_index], 'Hz')}"
        cbar.set_label(label, fontsize=10)
        ax.tick_params(axis="x", bottom=True, top=False)
        ax.tick_params(axis="y", left=True, right=False, rotation=90)
        ax.set_xlabel(rf"{self.frame_data['phi_long_name']}")
        ax.set_ylabel(rf"{self.frame_data['theta_long_name']}")

        ax.set_aspect("equal")

        return ax

    def plot(
        self,
        stokes: str = "I",
        nu_index: int = 0,
        t_index: int = 0,
        cmap: str = "default",
        rel_vmin: float = 0.005,
        rel_vmax: float = 0.995,
        filepath: str = None,
    ):
        X = np.r_[self.x_bins, self.y_bins]
        grid_u = Quantity(X, "rad").u
        grid_center = Quantity(self.center, "rad")

        header = fits.header.Header()
        header["CDELT1"] = -grid_u["factor"]
        header["CDELT2"] = grid_u["factor"]
        header["CRPIX1"] = 1
        header["CRPIX2"] = 1
        header["CTYPE1"] = "RA---SIN"
        header["CUNIT1"] = "deg     "
        header["CTYPE2"] = "DEC--SIN"
        header["CUNIT2"] = "deg     "
        header["RADESYS"] = "FK5     "
        header["CRVAL1"] = grid_center.deg[0]
        header["CRVAL2"] = grid_center.deg[1]

        # nu_index = 0
        # t_index = 0
        # stokes = [["I", "Q"],
        #         ["U", "V"]]
        # stokes = ["I", "Q", "U"]
        # stokes = [["I", "Q"]]
        NU_INDEX, T_INDEX, STOKES = [np.atleast_2d(x) for x in np.broadcast_arrays(nu_index, t_index, stokes)]

        nrows, ncols = STOKES.shape

        max_fig_size = 12

        ax_size = 5  # np.minimum(np.maximum(np.minimum(max_fig_size / nrows, max_fig_size / ncols), 5), 8)

        fig, axes = plt.subplots(
            nrows,
            ncols,
            figsize=(ax_size * ncols, ax_size * nrows),
            subplot_kw={"projection": WCS(header)},
            constrained_layout=True,
            sharex=True,
            sharey=True,
        )
        axes = np.atleast_1d(axes).reshape(STOKES.shape)

        for r in range(nrows):
            for c in range(ncols):
                ax = axes[r, c]

                if c > 0:
                    ax.coords[1].set_ticks_visible(False)
                    ax.coords[1].set_ticklabel_visible(False)
                # if c == ncols - 1:
                #     ay2 = ax.secondary_yaxis("right")
                #     ay2.set_ylabel(rf"$\Delta\,\theta_y$ [{grid_u['units']}]")
                if r == 0:
                    ax2 = ax.secondary_xaxis("top")
                    ax2.set_xlabel(rf"$\Delta\,\theta$ [{grid_u['units']}]")
                if r < nrows - 1:
                    ax.coords[0].set_ticks_visible(False)
                    ax.coords[0].set_ticklabel_visible(False)

                if STOKES[r, c] is None:
                    continue

                self.plot_slice(
                    fig,
                    ax,
                    stokes=STOKES[r, c],
                    nu_index=NU_INDEX[r, c],
                    t_index=T_INDEX[r, c],
                    rel_vmin=rel_vmin,
                    rel_vmax=rel_vmax,
                    cmap=cmap,
                )

                if filepath is not None:
                    plt.savefig(filepath=filepath, dpi=256)

    def to_fits(self, filepath):
        if self.dims.get("nu", np.nan) > 1:
            raise RuntimeError("Cannot write multifrequency maps to FITS")

        m = self.to(self.u["base_unit"])

        data = m.data[::-1, :]  # FITS counts from the bottom, while normal people count from the top
        if self.frame in ["ra_dec", "galatic"]:
            data = data[..., ::-1]

        fits.writeto(
            filename=filepath,
            data=data,
            header=m.header,
            overwrite=True,
        )

    def to_hdf(self, filename: str, compress: bool = True):
        compression_kwargs = {"compression": "gzip", "compression_opts": 9} if compress else {}

        with h5py.File(filename, "w") as f:
            f.create_dataset("data", dtype=np.float32, data=self.data, **compression_kwargs)
            if not (self.weight == 1).all().compute():
                f.create_dataset("weight", dtype=np.float32, data=self.weight, **compression_kwargs)
            for dim in ["stokes", "nu", "t"]:
                if dim in self.dims:
                    f.create_dataset(dim, data=getattr(self, dim))
            f.create_dataset("center", dtype=float, data=np.degrees(self.center))
            f.create_dataset("x_res", dtype=float, data=np.degrees(self.x_res))
            f.create_dataset("y_res", dtype=float, data=np.degrees(self.y_res))
            f.create_dataset("units", data=self.units)
            f.create_dataset("frame", data=self.frame)
            f.create_dataset("beam", data=np.degrees(self.beam))



from __future__ import annotations

import os

import numpy as np
import pandas as pd
import scipy as sp

from ..band import BandList
from ..beam import compute_angular_fwhm
from ..units import Quantity
from ..utils import compute_diameter

here, this_filename = os.path.split(__file__)


DET_COLUMN_TYPES = {
    "array": "str",
    "uid": "str",
    "array_name": "str",
    "band_name": "str",
    "band_center": "float",
    "sky_x": "float",
    "sky_y": "float",
    "baseline_x": "float",
    "baseline_y": "float",
    "baseline_z": "float",
    "polarized": "bool",
    "pol_angle": "float",
    "pol_label": "str",
    "primary_size": "float",
    "bath_temp": "float",
    "time_constant": "float",
    "white_noise": "float",
    "pink_noise": "float",
    "efficiency": "float",
}

SUPPORTED_ARRAY_PACKINGS = ["hex", "square", "sunflower"]
SUPPORTED_ARRAY_SHAPES = ["hex", "square", "circle"]


class ArrayList:
    def __init__(self, df: pd.DataFrame, bands: BandList, config: dict = {}):
        self.df = df
        self.df.index = np.arange(len(self.df.index))

        self.bands = BandList(bands)
        self.config = config

        for band_attr, det_attr in {
            "center": "band_center",
            "width": "band_width",
        }.items():
            self.df.loc[:, det_attr] = getattr(self, band_attr)

    def mask(self, **kwargs):
        mask = np.ones(len(self.df)).astype(bool)
        for k, v in kwargs.items():
            mask &= self.df.loc[:, k].values == v
        return mask

    def subset(self, **kwargs):
        return self._subset(self.mask(**kwargs))

    def _subset(self, mask):
        df = self.df.loc[mask]
        return ArrayList(
            df=df,
            bands=[b for b in self.bands if b.name in self.df.band_name.values],
        )

    def one_detector_from_each_band(self):
        first_det_mask = np.isin(
            np.arange(self.n),
            np.unique(self.band_name, return_index=True)[1],
        )
        return self._subset(mask=first_det_mask)

    def outer(self):
        outer_dets_index = sp.spatial.ConvexHull(self.offsets).vertices
        outer_dets_mask = np.isin(np.arange(self.n), outer_dets_index)
        return self._subset(mask=outer_dets_mask)

    @property
    def n(self):
        return len(self.df)

    @property
    def offsets(self):
        return np.c_[self.sky_x, self.sky_y]

    @property
    def baselines(self):
        return np.c_[self.baseline_x, self.baseline_y, self.baseline_z]

    @property
    def field_of_view(self):
        return Quantity(compute_diameter(self.offsets), "rad")

    @property
    def max_baseline(self):
        return Quantity(compute_diameter(self.baselines), "m")

    @property
    def index(self):
        return self.df.index.values

    @property
    def ubands(self):
        return list(self.bands.keys())

    @property
    def fwhm(self):
        """
        Returns the angular FWHM (in radians) at infinite distance.
        """
        return self.angular_fwhm(z=np.inf)

    def angular_fwhm(self, z):  # noqa F401
        """
        Angular beam width (in radians) as a function of depth (in meters)
        """
        nu = self.band_center  # in GHz
        return compute_angular_fwhm(z=z, fwhm_0=self.primary_size, n=1, nu=nu)

    def physical_fwhm(self, z):
        """
        Physical beam width (in meters) as a function of depth (in meters)
        """
        return z * self.angular_fwhm(z)

    def passband(self, nu):
        _nu = np.atleast_1d(nu)
        PB = np.zeros((len(self.df), len(_nu)))
        for band in self.bands:
            PB[self.band_name == band.name] = band.passband(_nu)
        return PB

    def __getattr__(self, attr):
        if attr in self.df.columns:
            return self.df.loc[:, attr].values.astype(DET_COLUMN_TYPES[attr])

        if all(hasattr(band, attr) for band in self.bands):
            values = np.zeros(shape=self.n, dtype=float)
            for band in self.bands:
                values[self.band_name == band.name] = getattr(band, attr)
            return values

        raise AttributeError(f"'{self.__class__.__name__}' object has no attribute '{attr}'")

    def __len__(self):
        return len(self.df)

    def __repr__(self):
        return self.df.__repr__()

    def _repr_html_(self):
        return self.df._repr_html_()



from .caching import *  # noqa
from .logging import *  # noqa

DEFAULT_TIME_FORMAT = "YYYY-MM-DD HH:mm:ss.SSS ZZ"


def leftpad(thing, n: int = 2, char=" "):
    return "\n".join([n * char + line for line in str(thing).splitlines()])



# for a given cross-section, extrude points orthogonally using a callback to compute covariance
from __future__ import annotations

import logging
from typing import Callable

import numpy as np
import pandas as pd
import scipy as sp
from pandas import DataFrame

logger = logging.getLogger("maria")


MIN_RES = {"2d": 5, "3d": 15}
MIN_RES_PER_BEAM = {"2d": 0.1, "3d": 0.5}
MIN_RES_PER_FOV = {"2d": 0.05, "3d": 0.1}

COV_MAT_JITTER = 1e-6


def generate_layers(
    sim,
    mode: str = "2d",
    angular: bool = True,
    max_height: float = 2e3,  # in meters
    min_res: float = None,  # in meters
    min_res_per_beam: float = None,
    min_res_per_fov: float = None,
    layer_spacing: float = 500,
) -> DataFrame:
    """
    Generate atmospheric layers.
    N.B.: We implicitly parametrize this model with a single, constant elevation.
    """

    min_res = min_res or MIN_RES[mode]
    min_res_per_beam = min_res_per_beam or MIN_RES_PER_BEAM[mode]
    min_res_per_fov = min_res_per_fov or MIN_RES_PER_FOV[mode]

    min_el = sim.boresight.el.min()

    h_samples = np.arange(0, max_height + 1e0, 1e-1)

    z_samples = h_samples / np.sin(min_el)

    fov = sim.instrument.dets.field_of_view.rad
    fwhm = sim.instrument.dets.one_detector_from_each_band().physical_fwhm(z_samples[:, None] + 1e-16).min(axis=1)
    r1 = min_res * np.ones(len(z_samples))
    r2 = min_res_per_beam * fwhm
    r3 = min_res_per_fov * z_samples * fov
    res_samples = np.minimum(1e3, np.maximum.reduce([r1, r2, r3]))

    def res_func(h):
        return sp.interpolate.interp1d(h_samples, res_samples)(h)

    if mode == "2d":
        h_boundaries = np.arange(0, max_height + layer_spacing, layer_spacing)
        process_index = np.arange(len(h_boundaries) - 1)
    elif mode == "3d":
        h_boundaries = [0]
        while True:
            new_h = h_boundaries[-1] + res_func(h_boundaries[-1])
            if new_h > max_height:
                break
            h_boundaries.append(h_boundaries[-1] + res_func(h_boundaries[-1]))
        h_boundaries = np.array(h_boundaries)
        process_index = 0

    h_centers = (h_boundaries[1:] + h_boundaries[:-1]) / 2

    weather = sim.atmosphere.weather

    weather_values = weather(altitude=sim.site.altitude + h_centers)

    layers = pd.DataFrame(weather_values)
    layers.insert(0, "process_index", process_index)
    layers.insert(1, "h", h_centers)
    layers.insert(2, "dh", np.diff(h_boundaries))
    layers.insert(3, "res", res_func(layers.h))
    layers.insert(4, "z", h_centers / np.sin(min_el))
    layers.insert(5, "angular", angular)

    h_boundaries = [0, *(layers.h.values[:-1] + layers.h.values[1:]) / 2, 1e5]

    for layer_index, (h1, h2) in enumerate(zip(h_boundaries[:-1], h_boundaries[1:])):
        dummy_h = sim.site.altitude + np.linspace(h1, h2, 1024)
        h = weather.altitude
        w = weather.absolute_humidity
        total_water = np.trapezoid(np.interp(dummy_h, h, w), x=dummy_h)
        layers.loc[layer_index, "total_water"] = total_water

    def boundary_layer_profile(h, h_0: float = 1e3, alpha: float = 1 / 7):
        return np.exp(-h / h_0) * h**alpha

    rel_var = boundary_layer_profile(layers.h.values) ** 2
    pwv_var = (weather.pwv * sim.atmosphere.pwv_rms_frac) ** 2 * rel_var / sum(rel_var)
    layers.loc[:, "pwv_rms"] = np.sqrt(pwv_var)

    if angular:
        layers.res /= layers.z

    return layers


def construct_extrusion_layers(
    points: float,
    res_func: Callable,
    z_min: float,
    z_max: float,
    mode: str = "3d",
    **mode_kwargs,
):
    triangulation = sp.spatial.Delaunay(points[..., 1:])
    layers = pd.DataFrame(columns=["x", "z", "n", "res", "indices"])

    layer_spacing = 500

    n = 0
    z = z_min if mode == "3d" else layer_spacing / 2

    while z < z_max:
        i = len(layers)

        res = res_func(z)

        # find a line
        wide_lp_x_dense = np.arange(points[..., 1].min(), points[..., 1].max(), 1e0)
        wide_lp_z_dense = z * np.ones(len(wide_lp_x_dense))
        wide_lp_dense = np.c_[wide_lp_x_dense, wide_lp_z_dense]
        interior = triangulation.find_simplex(wide_lp_dense) > -1
        lp_dense_x = wide_lp_x_dense[interior]
        n_lp = np.maximum(2, int(np.ptp(np.atleast_1d(lp_dense_x)) / res))
        lp_x = np.linspace(lp_dense_x.min() - 2 * res, lp_dense_x.max() + 2 * res, n_lp)

        layers.loc[i, "x"] = lp_x
        layers.loc[i, "z"] = z
        layers.loc[i, "n"] = n_lp
        layers.loc[i, "res"] = res
        layers.loc[i, "indices"] = n + np.arange(n_lp)

        z += res if mode == "3d" else layer_spacing
        n += n_lp

    cross_section_x = np.concatenate(layers.x.values)
    cross_section_z = np.concatenate(
        [entry.z * np.ones(entry.n) for index, entry in layers.iterrows()],
    )
    cross_section_points = np.concatenate(
        [cross_section_x[..., None], cross_section_z[..., None]],
        axis=-1,
    )

    extrusion_points = np.arange(
        points[..., 0].min() - 2 * layers.res.min(),
        points[..., 0].max() + 2 * layers.res.min(),
        layers.res.min(),
    )

    return layers, cross_section_points, extrusion_points



from __future__ import annotations

import os
from collections.abc import Sequence

import dask.array as da
import numpy as np
import scipy as sp
from tqdm import tqdm

from ..map import ProjectedMap
from ..tod import TOD
from .base import BaseMapper

np.seterr(invalid="ignore")

here, this_filename = os.path.split(__file__)


class BinMapper(BaseMapper):
    def __init__(
        self,
        center: tuple[float, float] = (0, 0),
        stokes: str = "I",
        width: float = 1,
        height: float = None,
        resolution: float = 0.01,
        frame: str = "ra_dec",
        units: str = "K_RJ",
        degrees: bool = True,
        calibrate: bool = False,
        tod_preprocessing: dict = {},
        map_postprocessing: dict = {},
        tods: Sequence[TOD] = [],
    ):
        height = height or width

        super().__init__(
            center=center,
            stokes=stokes,
            width=width,
            height=height,
            resolution=resolution,
            frame=frame,
            degrees=degrees,
            calibrate=calibrate,
            tods=tods,
            units=units,
        )

        self.tod_preprocessing = tod_preprocessing
        self.map_postprocessing = map_postprocessing

        self.n_x = int(np.maximum(1, self.width / self.resolution))
        self.n_y = int(np.maximum(1, self.height / self.resolution))

        self.x_bins = np.linspace(-0.5 * self.width, 0.5 * self.width, self.n_x + 1)
        self.y_bins = np.linspace(0.5 * self.height, -0.5 * self.height, self.n_y + 1)

    def _run(self, band):
        """
        The actual mapper for the BinMapper.
        """

        # nu = np.unique([band.center for tod in self.tods for band in tod.dets.bands])

        self.map = ProjectedMap(
            data=np.zeros((self.n_y, self.n_x)),
            width=self.width,
            center=self.center,
            frame=self.frame,
            degrees=False,
        )

        map_numer = np.zeros((self.n_stokes, (self.n_y + 2) * (self.n_x + 2)))
        map_denom = np.zeros((self.n_stokes, (self.n_y + 2) * (self.n_x + 2)))

        for tod in self.tods:
            band_tod = tod.subset(band=band.name)

            if not tod.shape[0] > 0:
                continue

            band_tod = band_tod.process(config=self.tod_preprocessing).to(self.units)

            P = self.map.pointing_matrix(coords=band_tod.coords)
            D = band_tod.signal.compute()
            W = band_tod.weight.compute()

            stokes_weight = band_tod.dets.stokes_weight()

            stokes_pbar = tqdm(enumerate(self.stokes), total=self.n_stokes, desc=f"Mapping band {band.name}")

            for stokes_index, stokes in stokes_pbar:
                stokes_pbar.set_postfix(band=band.name, stokes=stokes)

                s = stokes_weight[:, "IQUV".index(stokes)][..., None]

                map_numer[stokes_index] += P @ (np.sign(s) * W * D).ravel()
                map_denom[stokes_index] += P @ (np.abs(s) * W).ravel()

                # band_map_data["sum"][stokes_index] += sp.stats.binned_statistic_2d(
                #     dy.ravel(),
                #     dx.ravel(),
                #     (np.sign(w) * band_tod.weight * band_tod.signal).ravel(),
                #     bins=(self.y_bins[::-1], self.x_bins),
                #     statistic="sum",
                # ).statistic[::-1]

                # band_map_data["weight"][stokes_index] += sp.stats.binned_statistic_2d(
                #     dy.ravel(),
                #     dx.ravel(),
                #     (np.abs(w) * band_tod.weight).ravel(),
                #     bins=(self.y_bins[::-1], self.x_bins),
                #     statistic="sum",
                # ).statistic[::-1]

            del band_tod

        band_map_data = {
            "numer": map_numer.reshape(self.n_stokes, self.n_y + 2, self.n_x + 2)[:, 1:-1, 1:-1],
            "denom": map_denom.reshape(self.n_stokes, self.n_y + 2, self.n_x + 2)[:, 1:-1, 1:-1],
            "nom_freq": band.center,
        }

        return band_map_data



from __future__ import annotations

import logging
import os
from collections.abc import Mapping
from pathlib import Path
from typing import Union

import arrow
import jax.numpy as jnp
import numpy as np
import pandas as pd
import scipy as sp
from arrow import Arrow
from astropy.io import fits
from astropy.wcs import WCS
from matplotlib import pyplot as plt

from .. import coords
from ..coords import frames
from ..io import DEFAULT_TIME_FORMAT
from ..units import Quantity
from ..utils import compute_diameter, read_yaml, repr_phi_theta
from .patterns import get_scan_pattern_generator, scan_patterns

here, this_filename = os.path.split(__file__)
logger = logging.getLogger("maria")

all_patterns = list(scan_patterns.index.values)

MAX_VELOCITY_WARN = 10  # in deg/s
MAX_ACCELERATION_WARN = 10  # in deg/s

MIN_ELEVATION_WARN = 20  # in deg
MIN_ELEVATION_ERROR = 10  # in deg

here, this_filename = os.path.split(__file__)

PLAN_CONFIGS = {}
for plans_path in Path(f"{here}/plans").glob("*.yml"):
    PLAN_CONFIGS.update(read_yaml(plans_path))

plan_params = set()
for key, config in PLAN_CONFIGS.items():
    plan_params |= set(config.keys())
plan_data = pd.DataFrame(PLAN_CONFIGS).T
all_plans = list(plan_data.index.values)


class UnsupportedPlanError(Exception):
    def __init__(self, invalid_plan):
        super().__init__(
            f"The plan '{invalid_plan}' is not a supported plan. Supported plans are: \n\n{plan_data.to_string()}",
        )


def get_plan_config(plan_name="one_minute_zenith_stare", **kwargs):
    if plan_name not in PLAN_CONFIGS.keys():
        raise UnsupportedPlanError(plan_name)
    plan_config = PLAN_CONFIGS[plan_name].copy()
    for k, v in kwargs.items():
        plan_config[k] = v
    return plan_config


def get_plan(plan_name="one_minute_zenith_stare", **kwargs):
    plan_config = get_plan_config(plan_name, **kwargs)
    return Plan(**plan_config)


PLAN_FIELDS = {
    "start_time": Union[float, str, Arrow],
    "duration": float,
    "sample_rate": float,
    "frame": str,
    "degrees": bool,
    "scan_center": float,
    "scan_pattern": float,
    "scan_options": Mapping,
}


def validate_pointing_kwargs(kwargs):
    """
    Make sure that we have all the ingredients to produce the plan data.
    """
    if ("end_time" not in kwargs.keys()) and ("duration" not in kwargs.keys()):
        raise ValueError(
            """One of 'end_time' or 'duration' must be in the plan kwargs.""",
        )


class Plan:
    """
    A dataclass containing time-ordered plan data.
    """

    def __init__(
        self,
        description: str = "",
        start_time: str | int = "2024-02-10T06:00:00",
        duration: float = 60.0,
        sample_rate: float = 20.0,
        frame: str = "ra_dec",
        degrees: bool = True,
        jitter: float = 0,
        scan_center: tuple[float, float] = (4, 10.5),
        scan_pattern: str = "daisy",
        scan_options: dict = {},
    ):
        self.description = description
        self.start_time = start_time
        self.duration = duration
        self.sample_rate = sample_rate
        self.frame = frame
        self.degrees = degrees
        self.jitter = jitter
        self.scan_center = Quantity(scan_center, units=("deg" if degrees else "rad"))
        self.scan_pattern = scan_pattern
        self.scan_options = scan_options

        if not self.sample_rate > 0:
            raise ValueError("Parameter 'sample_rate' must be greater than zero!")

        # for k, v in PLAN_CONFIGS[self.scan_pattern]["scan_options"].items():
        #     if k not in self.scan_options.keys():
        #         self.scan_options[k] = v

        if not hasattr(self, "start_time"):
            self.start_time = arrow.now().timestamp()
        self.start_time = arrow.get(self.start_time)
        self.end_time = self.start_time.shift(seconds=self.duration)

        self.time_min = self.start_time.timestamp()
        self.time_max = self.end_time.timestamp()
        self.dt = 1 / self.sample_rate

        self.time = np.arange(self.time_min, self.time_max, self.dt)
        self.n_time = len(self.time)

        # # convert radius to width / height
        # if "width" in self.scan_options:
        #     self.scan_options["radius"] = 0.5 * self.scan_options.pop("width")

        # this is in pointing_units
        scan_offsets = get_scan_pattern_generator(self.scan_pattern)(
            self.time,
            **self.scan_options,
        )

        self.scan_offsets = Quantity(scan_offsets, units=("deg" if degrees else "rad")).rad

        scan_velocity = np.gradient(
            self.scan_offsets,
            axis=1,
            edge_order=0,
        ) / np.gradient(self.time)

        scan_acceleration = np.gradient(
            scan_velocity,
            axis=1,
            edge_order=0,
        ) / np.gradient(self.time)

        self.max_vel_deg = np.degrees(np.sqrt(np.sum(scan_velocity**2, axis=0)).max())
        self.max_acc_deg = np.degrees(np.sqrt(np.sum(scan_acceleration**2, axis=0)).max())

        if self.max_vel_deg > MAX_VELOCITY_WARN:
            logger.warning(
                (
                    f"The maximum velocity of the boresight ({self.max_vel_deg:.01f} deg/s) is "
                    "physically unrealistic. If this is undesired, double-check the parameters for your scan strategy."
                ),
                stacklevel=2,
            )

        if self.max_acc_deg > MAX_ACCELERATION_WARN:
            logger.warning(
                (
                    f"The maximum acceleration of the boresight ({self.max_acc_deg:.01f} deg/s^2) is "
                    "physically unrealistic. If this is undesired, double-check the parameters for your scan strategy."
                ),
                stacklevel=2,
            )

        self.scan_offsets += np.radians(self.jitter) * np.random.standard_normal(size=self.scan_offsets.shape)  # noqa

        pt = coords.offsets_to_phi_theta(
            self.scan_offsets.T,
            *self.scan_center.rad,
        )

        if self.frame == "ra_dec":
            self.ra = self.phi = pt[..., 0]
            self.dec = self.theta = pt[..., 1]
        elif self.frame == "az_el":
            self.az = self.phi = pt[..., 0]
            self.el = self.theta = pt[..., 1]
        else:
            raise ValueError("Not a valid pointing frame!")

    @property
    def frame_data(self):
        return frames[self.frame]

    def plot(self):
        q_offsets = Quantity(self.scan_offsets, units="rad")

        frame = coords.frames[self.frame]
        header = fits.header.Header()
        header["CDELT1"] = -q_offsets.u["factor"]
        header["CDELT2"] = q_offsets.u["factor"]
        header["CRPIX1"] = 1
        header["CRPIX2"] = 1
        header["CTYPE1"] = "RA---SIN"
        header["CUNIT1"] = "deg     "
        header["CTYPE2"] = "DEC--SIN"
        header["CUNIT2"] = "deg     "
        header["RADESYS"] = "FK5     "
        header["CRVAL1"] = self.scan_center.deg[0]
        header["CRVAL2"] = self.scan_center.deg[1]

        wcs = WCS(header)

        fig = plt.figure(figsize=(5, 5), dpi=256, constrained_layout=True)
        ax = fig.add_subplot(projection=wcs)

        cphi_repr, ctheta_repr = repr_phi_theta(*self.scan_center.rad, frame=self.frame)

        ax.plot(*q_offsets.value, lw=5e-1)
        ax.scatter(0, 0, c="r", marker="x", label=f"{cphi_repr}\n{ctheta_repr}")

        ax.legend(loc="upper right")

        ax.set_aspect("equal")

        ax.tick_params(axis="x", bottom=True, top=False)
        ax.tick_params(axis="y", left=True, right=False, rotation=90)
        ax2 = ax.secondary_xaxis("top")
        ay2 = ax.secondary_yaxis("right")
        ax2.set_xlabel(rf"$\Delta \, \theta_x$ [${q_offsets.u['math_name']}$]")
        ay2.set_ylabel(rf"$\Delta \, \theta_y$ [${q_offsets.u['math_name']}$]")
        ax.set_xlabel(rf"{self.frame_data['phi_long_name']}")
        ax.set_ylabel(rf"{self.frame_data['theta_long_name']}")

    def map_counts(self, instrument=None, x_bins=100, y_bins=100):
        array_offsets = np.zeros((1, 1, 2)) if instrument is None else instrument.offsets[:, None]

        OFFSETS = self.scan_offsets.T[None] + array_offsets

        xmin, ymin = OFFSETS.min(axis=(0, 1))
        xmax, ymax = OFFSETS.max(axis=(0, 1))

        if isinstance(x_bins, int):
            x_bins = np.linspace(xmin, max(xmax, xmin + 1e-6), x_bins + 1)
        if isinstance(y_bins, int):
            y_bins = np.linspace(ymin, max(ymax, ymin + 1e-6), y_bins + 1)

        bs = sp.stats.binned_statistic_2d(
            OFFSETS[..., 1].ravel(),
            OFFSETS[..., 0].ravel(),
            0,
            statistic="count",
            bins=(y_bins, x_bins),
        )

        return x_bins, y_bins, bs[0]

    def plot_counts(self, instrument=None, x_bins=100, y_bins=100):
        fig, ax = plt.subplots(1, 1, figsize=(5, 4))

        x, y, counts = self.map_counts(instrument=instrument, x_bins=x_bins, y_bins=y_bins)
        q_x = Quantity(x, "rad")
        q_y = Quantity(y, "rad")

        heatmap = ax.pcolormesh(q_x.value, q_y.value, counts, cmap="turbo", vmin=0)
        ax.set_xlabel(rf"$\Delta \theta_x$ [{q_x.u['math_name']}]")
        ax.set_ylabel(rf"$\Delta \theta_y$ [{q_y.u['math_name']}]")

        cbar = fig.colorbar(heatmap, location="right")
        cbar.set_label("counts")

    def __repr__(self):
        cphi_repr, ctheta_repr = repr_phi_theta(*self.scan_center.rad, frame=self.frame)

        return f"""Plan:
  start_time: {self.start_time.format(DEFAULT_TIME_FORMAT)}
  duration: {Quantity(self.duration, "s")}
  sample_rate: {Quantity(self.sample_rate, "Hz")}
  center:
    {cphi_repr}
    {ctheta_repr}
  scan_pattern: {self.scan_pattern}
  scan_radius: {Quantity(compute_diameter(self.scan_offsets.T), "rad")}
  scan_kwargs: {self.scan_options}"""



from __future__ import annotations

import warnings
from itertools import cycle

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
import scipy as sp
from matplotlib.animation import FuncAnimation
from matplotlib.collections import EllipseCollection
from matplotlib.patches import Patch

from ..beam import compute_angular_fwhm
from ..units import Quantity
from ..utils.signal import detrend as detrend_signal
from ..utils.signal import remove_slope


def tod_plot(
    tod,
    detrend: bool = True,
    n_freq_bins: int = 1024,
    max_dets: int = 100,
    lw: float = 1e0,
    fontsize: float = 8,
):
    fig, axes = plt.subplots(
        ncols=2,
        nrows=len(tod.fields),
        sharex="col",
        figsize=(10, 4),
        dpi=256,
    )
    axes = np.atleast_2d(axes)
    gs = axes[0, 0].get_gridspec()

    plt.subplots_adjust(top=0.99, bottom=0.01, hspace=0, wspace=0.02)

    for ax in axes[:, -1]:
        ax.remove()
    ps_ax = fig.add_subplot(gs[:, -1])

    handles = []
    colors = cycle(plt.rcParams["axes.prop_cycle"].by_key()["color"])

    tod_ubands = np.unique(tod.dets.band_name)
    power_spectra = np.zeros((len(tod_ubands), len(tod.fields), n_freq_bins - 1))

    for field_index, field in enumerate(tod.fields):
        data = tod.data[field]

        tod_ax = axes[field_index, 0]

        for band_index, band_name in enumerate(tod_ubands):
            color = next(colors)

            band_mask = tod.dets.band_name == band_name
            signal = data[band_mask]

            if band_mask.sum() > max_dets:
                signal = signal[np.random.choice(np.arange(len(signal)), max_dets, replace=False)]

            if detrend:
                signal = detrend_signal(signal, order=1)

            f, ps = sp.signal.periodogram(remove_slope(signal.compute()), fs=tod.fs, window="tukey")

            f_bins = np.geomspace(f[1], f[-1], n_freq_bins)
            f_mids = np.sqrt(f_bins[1:] * f_bins[:-1])

            binned_ps = sp.stats.binned_statistic(
                f,
                ps.mean(axis=0),
                bins=f_bins,
                statistic="mean",
            )[0]

            power_spectra[band_index, field_index] = binned_ps

            use = binned_ps > 0

            ps_ax.plot(
                f_mids[use],
                binned_ps[use],
                lw=lw,
                color=color,
            )
            tod_ax.plot(
                tod.time,
                signal.mean(axis=0),
                lw=lw,
                color=color,
            )

            handles.append(Patch(color=color, label=f"{field} ({band_name})"))

        tod_ax.set_xlim(tod.time.min(), tod.time.max())

        label = tod_ax.text(
            0.01,
            0.99,
            f"{field} [{tod.units}]",
            fontsize=fontsize,
            ha="left",
            va="top",
            transform=tod_ax.transAxes,
        )
        label.set_bbox(dict(facecolor="white", alpha=0.8))

    tod_ax.set_xlabel("Time [s]", fontsize=fontsize)

    ps_ax.yaxis.tick_right()
    ps_ax.yaxis.set_label_position("right")
    ps_ax.set_xlabel("Frequency [Hz]", fontsize=fontsize)
    ps_ax.set_ylabel(f"[{tod.units}$^2$/Hz]", fontsize=fontsize)
    ps_ax.legend(handles=handles, loc="upper right", fontsize=fontsize)
    ps_ax.loglog()
    ps_ax.set_xlim(f_mids.min(), f_mids.max())

    with warnings.catch_warnings():
        warnings.simplefilter("ignore")
        minimum_dominant_ps = np.nanmin(np.nanmax(power_spectra, axis=1))
        minimum_maximum_ps = np.nanmin(np.nanmax(power_spectra, axis=2))

    ps_ax.set_ylim(1e-1 * np.minimum(minimum_dominant_ps, minimum_maximum_ps))


def twinkle_plot(tod, rate=2, fps=30, start_index=0, max_frames=100, filename=None):
    fps = np.minimum(fps, tod.fs)
    time_offset = tod.time.min()

    frame_time = np.arange(tod.time.min(), tod.time.max(), rate / fps)[:max_frames]
    frame_index = np.interp(frame_time, tod.time, np.arange(len(tod.time))).astype(int)

    offsets = Quantity(np.c_[tod.dets.sky_x, tod.dets.sky_y], "rad")
    fwhms = Quantity(compute_angular_fwhm(fwhm_0=tod.dets.primary_size, nu=tod.dets.band_center), "rad")

    bands = sorted(np.unique(tod.dets.band_name))

    nrows = int(np.sqrt(len(bands)))
    ncols = int(np.ceil(len(bands) / nrows))
    subplot_width = np.minimum(6, 10 / nrows)
    subplot_height = np.minimum(6, 10 / ncols)
    subplot_size = np.minimum(subplot_width, subplot_height)
    figure_size = (ncols * subplot_size, nrows * (subplot_size + 1))

    fig, axes = plt.subplots(
        nrows=nrows,
        ncols=ncols,
        sharex=True,
        sharey=True,
        figsize=figure_size,
        dpi=256,
        constrained_layout=True,
    )

    axes_iterator = iter(np.atleast_1d(axes).ravel())

    subplots = {}

    for band in bands:
        ax = next(axes_iterator)

        band_mask = tod.dets.band_name == band

        ec = EllipseCollection(
            widths=getattr(fwhms, offsets.units)[band_mask],
            heights=getattr(fwhms, offsets.units)[band_mask],
            angles=0,
            units="xy",
            offsets=getattr(offsets, offsets.units)[band_mask],
            transOffset=ax.transData,
            cmap="cmb",
        )

        ax.add_collection(ec)
        ax.scatter(
            *getattr(offsets, offsets.units)[band_mask].T,
            label=band,
            s=0,
            color="k",
        )

        ax.set_xlabel(r"$\Delta \theta_x$ [deg.]")
        ax.set_ylabel(r"$\Delta \theta_y$ [deg.]")

        boresight_info = ax.text(
            0.01,
            0.99,
            "",
            ha="left",
            va="top",
            transform=ax.transAxes,
        )

        band_qdata = Quantity(tod.signal[band_mask][..., frame_index].compute(), units=tod.units)

        cbar = fig.colorbar(ec, ax=ax, shrink=0.8, location="bottom")

        cbar.set_label(f"{band_qdata.q['long_name']} [${band_qdata.u['math_name']}$]")

        subplots[band] = {
            "ax": ax,
            "ec": ec,
            "cbar": cbar,
            "info": boresight_info,
            "time": tod.time[frame_index] - time_offset,
            "az": tod.boresight.az[frame_index],
            "el": tod.boresight.el[frame_index],
            "data": band_qdata.value,
        }

    for ax in axes_iterator:
        ax.set_axis_off()

    def update(frame, subplots):
        # i = frame_index[frame]

        for band, subplot in subplots.items():
            info = f"time = {time_offset:.00f} + {subplot['time'][frame]:.02f} s\naz = {np.degrees(subplot['az'][frame]):.02f} deg\nel = {np.degrees(subplot['el'][frame]):.02f} deg"  # noqa

            buffer = 4

            norm_start = np.maximum(0, frame - buffer)
            norm_end = np.minimum(len(frame_index) - 1, frame + buffer)

            # if frame < buffer:
            #     norm_start, norm_end = 0, 2 * buffer
            # if frame > len(t) - buffer:
            #     norm_start, norm_end = len(t) - 2 * buffer - 1, len(t) - 1

            norm = mpl.colors.Normalize(
                *np.quantile(subplot["data"][:, norm_start:norm_end], q=[0.01, 0.99]),
            )

            subplot["ec"].set_array(subplot["data"][:, frame])
            subplot["ec"].set_norm(norm)
            subplot["cbar"].update_normal(subplot["ec"])
            subplot["info"].set_text(info)  # noqa

        return subplots

    update(0, subplots)

    ani = FuncAnimation(
        fig=fig,
        func=update,
        fargs=(subplots,),
        frames=len(frame_index),
        interval=1e3 / fps,
    )

    if filename:
        ani.save(filename)



from __future__ import annotations

import logging
import os
import time as ttime

import dask.array as da
import numpy as np
import scipy as sp
from jax import scipy as jsp
from tqdm import tqdm

from maria.io import humanize_time

from ..constants import k_B

here, this_filename = os.path.split(__file__)

logger = logging.getLogger("maria")


class AtmosphereMixin:
    def _simulate_atmosphere(self):
        # this produces self.atmosphere.zenith_scaled_pwv at lower res
        # which we use to compute emission and opacity
        self.atmosphere.simulate_pwv()

        # update to the sim resolution
        self.zenith_scaled_pwv = da.from_array(
            sp.interpolate.interp1d(
                self.atmosphere.coords.t,
                self.atmosphere.zenith_scaled_pwv,
                bounds_error=False,
                fill_value="extrapolate",
            )(self.coords.t),
        )

    def _compute_atmospheric_emission(self):
        atmosphere_loading = np.zeros(self.zenith_scaled_pwv.shape, dtype=self.dtype)

        bands_pbar = tqdm(
            self.instrument.dets.bands,
            desc="Computing atmospheric emission",
            disable=self.disable_progress_bars,
        )
        for band in bands_pbar:
            emission_s = ttime.monotonic()

            bands_pbar.set_postfix({"band": band.name})

            band_index = self.instrument.dets.mask(band_name=band.name)

            # the 1e12 is for picowatts
            det_power_grid = (
                1e12
                * k_B
                * np.trapezoid(
                    self.atmosphere.spectrum._emission * band.passband(self.atmosphere.spectrum.side_nu),
                    self.atmosphere.spectrum.side_nu,
                    axis=-1,
                )
            )

            band_power_interpolator = jsp.interpolate.RegularGridInterpolator(
                self.atmosphere.spectrum.points[:3],
                det_power_grid,
            )

            atmosphere_loading[band_index] = band_power_interpolator(
                (
                    self.atmosphere.weather.temperature[0],
                    self.zenith_scaled_pwv[band_index].compute(),
                    self.coords.el[band_index].clip(max=np.pi / 2),
                ),
            )

            logger.debug(
                f"Computed atmospheric emission for band {band.name} in {humanize_time(ttime.monotonic() - emission_s)}."
            )

        self.loading["atmosphere"] = da.asarray(atmosphere_loading, dtype=self.dtype)
        del atmosphere_loading



from __future__ import annotations

import logging
import os
import time as ttime
import warnings
from collections.abc import Sequence

import numpy as np
import scipy as sp

from ..instrument import BandList
from ..io import humanize_time
from ..map import ProjectedMap
from ..tod import TOD
from ..units import Quantity

# np.seterr(invalid="ignore")

here, this_filename = os.path.split(__file__)
logger = logging.getLogger("maria")


class BaseMapper:
    """
    The base class for mapping.
    """

    def __init__(
        self,
        center: tuple[float, float],
        stokes: float,
        width: float,
        height: float,
        resolution: float,
        frame: str,
        units: str,
        degrees: bool,
        calibrate: bool,
        tods: Sequence[TOD],
    ):
        self.resolution = np.radians(resolution) if degrees else resolution
        self.center = np.radians(center) if degrees else center
        self.width = np.radians(width) if degrees else width
        self.height = np.radians(height) if degrees else height
        self.degrees = degrees
        self.calibrate = calibrate
        self.frame = frame
        self.units = units
        self.stokes = stokes

        self.bands = BandList(bands=[])

        self.tods = []
        self.add_tods(tods)

    def plot(self):
        if not hasattr(self, "map"):
            raise RuntimeError("Mapper has not been run yet.")
        self.map.plot()

    @property
    def n_stokes(self):
        return len(self.stokes)

    def add_tods(self, tods):
        for tod in np.atleast_1d(tods):
            self.tods.append(tod)

            for band in tod.dets.bands:
                self.bands.add(band)

        # self.bands = list(
        #     np.unique([list(np.unique(tod.dets.band_name)) for tod in self.tods]),
        # )

    def _run(self):
        raise ValueError("Not implemented!")

    def run(self):
        if not len(self.tods):
            raise RuntimeError("This mapper has no TODs!")

        self.map_data = {}

        map_numer = np.zeros((self.n_stokes, len(self.bands), self.n_y, self.n_x))
        map_denom = np.zeros((self.n_stokes, len(self.bands), self.n_y, self.n_x))

        for band_index, band in enumerate(self.bands):
            band_start_s = ttime.monotonic()
            band_maps = self._run(band)

            map_numer[:, band_index] = band_maps["numer"]
            map_denom[:, band_index] = band_maps["denom"]

            logger.info(f"Ran mapper for band {band.name} in {humanize_time(ttime.monotonic() - band_start_s)}.")

        if "gaussian_filter" in self.map_postprocessing.keys():
            sigma = self.map_postprocessing["gaussian_filter"]["sigma"]

            map_numer = sp.ndimage.gaussian_filter(map_numer, sigma=(0, 0, sigma, sigma))
            map_denom = sp.ndimage.gaussian_filter(map_denom, sigma=(0, 0, sigma, sigma))

        if "median_filter" in self.map_postprocessing.keys():
            size = self.map_postprocessing["median_filter"]["size"]
            map_numer = sp.ndimage.median_filter(
                map_numer,
                size=size,
                axes=(-2, -1),
            )
            map_denom = sp.ndimage.median_filter(
                map_denom,
                size=size,
                axes=(-2, -1),
            )

        for stokes_index, stokes in enumerate(self.stokes):
            for nu_index, nu in enumerate(self.bands.center):
                if map_denom[stokes_index, nu_index].sum() == 0:
                    logger.warning(f"No counts for map (stokes={stokes}, nu={Quantity(nu, 'Hz')})")

        map_data = map_numer / map_denom

        # by convention maps have zero mean
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category=RuntimeWarning)
            map_offsets = np.nanmean(map_data, axis=(-1, -2))[..., None, None]

        return ProjectedMap(
            data=(map_data - map_offsets),
            stokes=self.stokes,
            weight=map_denom,
            nu=self.bands.center,
            resolution=self.resolution,
            center=self.center,
            degrees=False,
            frame=self.frame,
            units=self.units,
        )



from __future__ import annotations

import os
from pathlib import Path

import h5py
import healpy as hp
import matplotlib as mpl
import numpy as np
import pandas as pd
from astropy.coordinates import EarthLocation
from matplotlib import pyplot as plt

from ..io import fetch
from ..units import Quantity
from ..utils import read_yaml, repr_lat_lon

here, this_filename = os.path.split(__file__)

SITE_CONFIGS = {}
for sites_path in Path(f"{here}/sites").glob("*.yml"):
    SITE_CONFIGS.update(read_yaml(sites_path))

for site in SITE_CONFIGS:
    config = SITE_CONFIGS[site]
    SITE_CONFIGS[site]["instruments"] = ", ".join(config["instruments"]) if "instruments" in config else ""

SITE_DISPLAY_COLUMNS = [
    "description",
    "instruments",
    "region",
    "latitude",
    "longitude",
    "altitude",
]
site_data = pd.DataFrame(SITE_CONFIGS).T.sort_values("region")
all_sites = list(site_data.index.values)

REGION_DISPLAY_COLUMNS = ["location", "country", "latitude", "longitude"]
REGIONS = pd.read_csv(f"{here}/regions.csv", index_col=0)
all_regions = list(REGIONS.index.values)


def get_height_map():
    with h5py.File(fetch("world_heightmap.h5"), "r") as f:
        height_map = f["data"][:].astype(np.uint16)
    return 32 * np.where(height_map < 255, height_map, np.nan)


class InvalidRegionError(Exception):
    def __init__(self, invalid_region):
        super().__init__(
            f"The region '{invalid_region}' is not supported. "
            f"Supported regions are:\n\n{REGIONS.loc[:, REGION_DISPLAY_COLUMNS].to_string()}",
        )


class InvalidSiteError(Exception):
    def __init__(self, invalid_site):
        super().__init__(
            f"The site '{invalid_site}' is not supported. "
            f"Supported sites are:\n\n{site_data.loc[:, SITE_DISPLAY_COLUMNS].to_string()}",
        )


def get_location(site_name):
    site = get_site(site_name)
    return EarthLocation.from_geodetic(
        lon=site.longitude,
        lat=site.latitude,
        height=site.altitude,
    )


def get_site_config(site_name="hoagie_haven", **kwargs):
    if site_name not in SITE_CONFIGS.keys():
        raise InvalidSiteError(site_name)
    SITE_CONFIG = SITE_CONFIGS[site_name].copy()
    for k, v in kwargs.items():
        SITE_CONFIG[k] = v
    return SITE_CONFIG


def get_site(site_name="hoagie_haven", **kwargs):
    return Site(**get_site_config(site_name=site_name, **kwargs))


class Site:
    """
    An object representing a point on the earth that parametrizes observing conditions.
    """

    def __init__(
        self,
        description: str = "",
        region: str = "princeton",
        altitude: float = None,  # in meters
        seasonal: bool = True,
        diurnal: bool = True,
        latitude: float = None,  # in degrees
        longitude: float = None,  # in degrees
        weather_quantiles: dict = {},
        documentation: str = "",
        instruments: list = [],
    ):
        self.description = description
        self.region = region
        self.altitude = altitude
        self.seasonal = seasonal
        self.diurnal = diurnal
        self.latitude = latitude
        self.longitude = longitude
        self.weather_quantiles = weather_quantiles
        self.documentation = documentation
        self.instruments = instruments

        if self.region not in REGIONS.index.values:
            raise InvalidRegionError(self.region)

        if self.longitude is None:
            self.longitude = float(REGIONS.loc[self.region].longitude)

        if self.latitude is None:
            self.latitude = float(REGIONS.loc[self.region].latitude)

        if self.altitude is None:
            self.altitude = float(REGIONS.loc[self.region].altitude)

        self.earth_location = EarthLocation.from_geodetic(
            lon=self.longitude,
            lat=self.latitude,
            height=self.altitude,
        )

    def plot(self, res=0.03, wide_size: float = 45, zoom_size: float = 2):
        height_map = get_height_map()

        kwargs = {
            "rot": (self.longitude, self.latitude),
            "cmap": "gist_earth",
            "flip": "geo",
            "badcolor": "gray",
            "reso": 60 * res,
            "return_projected_map": True,
            "no_plot": True,
        }

        zoom_map = hp.gnomview(
            height_map,
            xsize=zoom_size / res,
            **kwargs,
        )  # , norm=mpl.colors.Normalize(vmin=0, vmax=5e3))
        wide_map = hp.gnomview(
            height_map,
            xsize=wide_size / res,
            **kwargs,
        )  # , norm=mpl.colors.Normalize(vmin=0, vmax=5e3))

        fig, axes = plt.subplots(1, 2, figsize=(8, 5), constrained_layout=True)

        zoom_vmax = np.nanpercentile(zoom_map.data, q=99)
        wide_vmax = np.nanpercentile(wide_map.data, q=99)

        cmap = mpl.cm.gist_ncar
        cmap = mpl.cm.gist_earth
        cmap.set_bad("royalblue", 1.0)

        wide_mesh = axes[0].pcolormesh(
            wide_size * np.linspace(-0.5, 0.5, int(wide_size / res)),
            wide_size * np.linspace(-0.5, 0.5, int(wide_size / res)),
            wide_map,
            cmap=cmap,
            vmax=wide_vmax,
        )
        cbar = fig.colorbar(wide_mesh, ax=axes[0], location="bottom", shrink=0.8, label="height")
        cbar.set_label("height [m]")

        zoom_mesh = axes[1].pcolormesh(
            zoom_size * np.linspace(-0.5, 0.5, int(zoom_size / res)),
            zoom_size * np.linspace(-0.5, 0.5, int(zoom_size / res)),
            zoom_map,
            cmap=cmap,
            vmax=zoom_vmax,
        )
        cbar = fig.colorbar(zoom_mesh, ax=axes[1], location="bottom", shrink=0.8, label="height")
        cbar.set_label("height [m]")

        for ax in axes:
            # ax.scatter(0, 0, color="gray", alpha=0.8, s=256)
            ax.scatter(0, 0, edgecolor="r", facecolor="none", linewidth=2, s=256)
            ax.set_xticks([])
            ax.set_yticks([])
            ax.set_facecolor("gray")
            ax.set_rasterized(True)

    def __repr__(self):
        s = f"""Site:
  region: {self.region}
  location: {repr_lat_lon(self.latitude, self.longitude)}
  altitude: {Quantity(self.altitude, "m")}
  seasonal: {self.seasonal}
  diurnal: {self.diurnal}"""

        return s

        # parts = {
        #     "location": f"({repr_lat_lon(self.latitude, self.longitude)})",
        #     "region": self.region,
        #     "altitude": f"{self.altitude}m",
        #     "seasonal": self.seasonal,
        #     "diurnal": self.diurnal,
        #     "weather_quantiles": self.weather_quantiles,
        # }

        # return rf"Site({', '.join([f'{k}={v}' for k, v in parts.items()])})"



from __future__ import annotations

import pytest

import maria
from maria.units import Quantity


def test_units():
    assert str(Quantity(0.1, "deg")) == "6’"
    assert str(Quantity(1e-8, "deg")) == "36 uarcsec"
    assert str(Quantity(1e6, "uK_RJ")) == "1 K_RJ"
    assert str(Quantity(1e15, "K_RJ")) == "1e+15 K_RJ"  # don't go high
    assert str(Quantity(1e-6, "Hz")) == "1e-06 Hz"  # don't go low
    assert str(Quantity(1000, "s")) == "1000 s"



import logging
import os
import time as ttime
from collections.abc import Iterable

import arrow
import dask.array as da
import numpy as np
import scipy as sp

from ..calibration import Calibration
from ..constants import MARIA_MAX_NU, MARIA_MIN_NU
from ..errors import FrequencyOutOfBoundsError
from ..units import Quantity, parse_units

logger = logging.getLogger("maria")

here, this_filename = os.path.split(__file__)

SLICE_DIMS = {
    "stokes": {
        "dtype": str,
        "default": "I",
    },
    "nu": {
        "dtype": float,
        "default": 150e9,
    },
    "t": {
        "dtype": float,
        "default": arrow.now().timestamp(),
    },
}


class Map:
    """
    The base class for maps. Maps have data
    """

    def __init__(
        self,
        data: float,
        weight: float,
        stokes: str,
        nu: Iterable[float],
        t: Iterable[float],
        z: Iterable[float],
        beam: tuple[float, float, float],  # noqa
        map_dims: dict,
        units: str = "K_RJ",
        degrees: bool = True,  # noqa
        dtype: type = np.float32,
    ):
        self.data = da.asarray(data).astype(dtype)
        self.weight = (da.asarray(weight) if weight is not None else da.ones_like(self.data)).astype(dtype)
        self.data, self.weight = np.broadcast_arrays(self.data, self.weight)

        slice_dims = {}
        if stokes is not None:
            if stokes not in ["I", "IQU", "IQUV"]:
                raise ValueError("'stokes' parameter must be either 'I', 'IQU', or 'IQUV'")
            self.stokes = stokes.upper()
            slice_dims["stokes"] = len(self.stokes)
        else:
            self.stokes = ""

        if nu is not None:
            self.nu = np.atleast_1d(nu)
            bad_freqs = list(self.nu[(self.nu < MARIA_MIN_NU) | (self.nu > MARIA_MAX_NU)])
            if bad_freqs:
                raise FrequencyOutOfBoundsError(bad_freqs)
            slice_dims["nu"] = len(self.nu)
        else:
            self.nu = np.array([])

        if t is not None:
            self.t = np.atleast_1d(t)
            slice_dims["t"] = len(self.t)
        else:
            self.t = np.array([])

        if z is not None:
            self.z = np.atleast_1d(z)
            slice_dims["z"] = len(self.z)
        else:
            self.z = np.array([])

        self.dims = {**slice_dims, **map_dims}

        self.data *= np.ones(list(self.dims.values()))
        self.weight *= np.ones(list(self.dims.values()))

        self.units = parse_units(units)["units"]

        # for i, (dim, n) in enumerate(slice_dims.items()):
        #     if self.data.shape[i] != n:
        #         raise ValueError(
        #             f"'{dim}' has length {n} but map has {dim} length {self.data.shape[i]}.",
        #         )

        implied_shape = tuple(list(self.dims.values()))
        # if len(implied_shape) != self.data.ndim:
        #     raise ValueError(f"Inputs imply rank {len(self.dims)} for map data, but data has rank {self.data.ndim}")

        if implied_shape != self.data.shape:
            raise ValueError(
                f"Inputs imply shape {self.dims_string}={implied_shape} for map data, but data has shape {self.data.shape}"
            )

    @property
    def ndim(self):
        return len(self.dims)

    @property
    def dims_string(self):
        return f"({', '.join(self.dims.keys())})"

    @property
    def dims_list(self):
        return list(self.dims.keys())

    @property
    def nu_bins(self):
        """
        This might break in the year 3000.
        """
        return np.array([0, *(self.nu[1:] + self.nu[:-1]), 1e6])

    @property
    def nu_side(self):
        return (self.nu_bins[:-1] + self.nu_bins[1:]) / 2

    @property
    def t_bins(self):
        """
        This might break in the year 3000.
        """
        t_min = arrow.get("0001-01-01").timestamp()
        t_max = arrow.get("3000-01-01").timestamp()

        return np.array([t_min, *(self.t[1:] + self.t[:-1]), t_max])

    @property
    def t_side(self):
        return (self.t_bins[:-1] + self.t_bins[1:]) / 2

    @property
    def u(self):
        return parse_units(self.units)

    # @property
    # def weight(self):
    #     return self._weight if self._weight is not None else da.ones(shape=self.data.shape)

    def squeeze(self, dim):
        package = self.package()
        for i, (name, n) in enumerate(self.dims.items()):
            if (name == dim) or (i == dim):
                if n == 1:
                    package["data"] = package["data"].squeeze(i)
                    package["weight"] = package["weight"].squeeze(i)
                    package.pop(name)
                    return type(self)(**package)
                else:
                    raise ValueError(f"Cannot squeeze dimension '{dim}' with length {n} > 1")
        raise ValueError(f"{self.__class__.__name__} has no dimension '{dim}'")

    def unsqueeze(self, dim, value=None):
        if dim not in SLICE_DIMS:
            raise ValueError(f"'{dim}' is not a valid map dimension")
        if dim in self.dims:
            raise Exception(f"{self.__class__.__name__} already has dimension '{dim}'")

        new_dim_index = 0
        for d in ["stokes", "nu", "t"]:
            if d in self.dims and d != dim:
                new_dim_index += 1

        package = self.package()
        package["data"] = np.expand_dims(package["data"], new_dim_index)
        package["weight"] = np.expand_dims(package["weight"], new_dim_index)
        if value is None:
            value = SLICE_DIMS[dim]["default"]
        package[dim] = value

        return type(self)(**package)

    @property
    def pixel_area(self):
        if hasattr(self, "resolution"):
            return self.resolution**2
        else:
            return self.x_res * self.y_res

    @property
    def beam_area(self):
        return np.pi * self.beam[0] * self.beam[1]

    def to(self, units: str):
        if units == self.units:
            return self

        u = parse_units(units)

        package = self.package().copy()

        # this is just a scaling by some factor
        if u["quantity"] == self.u["quantity"]:
            package["data"] *= self.u["factor"] / u["factor"]

        else:
            if "nu" not in self.dims:
                raise ValueError(
                    f"Cannot convert from quantity {self.u['quantity']} to quantity {u['quantity']} "
                    f"when map has no frequency."
                )

            data = package["data"].swapaxes(0, self.dims_list.index("nu"))  # put the nu index in front
            for nu_index, nu in enumerate(self.nu):
                cal = Calibration(
                    f"{self.units} -> {units}",
                    nu=nu,
                    pixel_area=self.pixel_area,
                    beam_area=self.beam_area,
                )
                data[nu_index] = cal(data[nu_index])

            package["data"] = data.swapaxes(0, self.dims_list.index("nu"))  # swap the axes back

        package["units"] = units
        return type(self)(**package)

    def sample_nu(self, nu):
        map_nu_interpolator = sp.interpolate.interp1d(self.nu, self.data, axis=1, kind="linear")

        nu_maps = []
        for nu in np.atleast_1d(nu):
            if nu < self.nu[0]:
                nu_maps.append(self.data[:, 0])
            elif not (nu < self.nu[-1]):  # this will include nan
                nu_maps.append(self.data[:, -1])
            else:
                nu_maps.append(map_nu_interpolator(nu))

        return np.stack(nu_maps, axis=1)

    @property
    def nu_bin_bounds(self):
        nu_boundaries = [0, *(self.nu[:-1] + self.nu[1:]) / 2, np.inf]
        return [(nu1, nu2) for nu1, nu2 in zip(nu_boundaries[:-1], nu_boundaries[1:])]



from __future__ import annotations

import numpy as np
import scipy as sp

from .filters import bandpass, highpass, lowpass  # noqa


def get_kernel(n, kind="triangle"):
    if kind == "triangle":
        kernel = np.r_[np.linspace(0, 1, n + 1)[1:], np.linspace(1, 0, n + 1)[1:-1]]

        return kernel / kernel.sum()


def fast_downsample(DATA, r=1):
    *in_shape, n = DATA.shape
    CUMSUM = np.cumsum(np.atleast_2d(DATA), axis=-1)
    return (CUMSUM[..., r::r] - CUMSUM[..., :-r:r]).reshape(*in_shape, -1) / r


def downsample(DATA, rate, axis=-1, method=None):
    if method is None:
        return np.swapaxes(np.swapaxes(DATA, 0, axis)[::rate], 0, axis)

    if method == "fast":
        return fast_downsample(
            DATA,
            r=rate,
        )

    if method == "flat":
        _DATA = np.swapaxes(DATA, 0, axis)
        cs_data = np.cumsum(_DATA, axis=0)
        return np.swapaxes((cs_data[rate::rate] - cs_data[:-rate:rate]) / rate, 0, axis)

    if method == "triangle":
        if rate == 1:
            return DATA
        if rate < 1:
            raise ValueError("downsample rate must be an integer >= 1")

        _DATA = np.swapaxes(DATA, 0, axis)
        kernel = np.expand_dims(
            get_kernel(n=rate, kind="triangle"),
            axis=tuple(np.arange(1, len(DATA.shape))),
        )
        n_kern = len(kernel)
        starts = np.arange(0, len(_DATA) - n_kern, rate)
        ends = starts + n_kern

        return np.swapaxes(
            np.r_[[np.sum(_DATA[s:e] * kernel, axis=0) for s, e in zip(starts, ends)]],
            0,
            axis,
        )


def decompose(data, k: int = None):
    k = k or len(data)

    if data.ndim != 2:
        raise ValueError("'data' must be a 2D array.")

    if k < len(data):
        u, s, vh = sp.sparse.linalg.svds(data, k=k, which="LM")
    else:
        u, s, vh = sp.linalg.svd(data, full_matrices=False)

    norm = np.sign(u.mean(axis=0)) * vh.std(axis=-1)
    a = u @ np.diag(s * norm)
    b = np.diag(1 / norm) @ vh

    mode_sort = np.argsort(-s)

    return a[:, mode_sort], b[mode_sort]


def spline(x: float, y: float, spacing: float, order: int = 3):
    # this has shape (n_basis, len(x))
    B = bspline_basis(x, spacing=spacing, order=order)

    A = B @ B.T


def bspline_basis(x, spacing, order=3):
    k = np.arange(np.min(x), np.max(x), spacing)
    if not len(k) > 0:
        k = np.array([np.mean(x)])
    k += x.mean() - k.mean()
    t = np.r_[
        k[0] + spacing * np.arange(-order - 1, 0),
        k,
        k[-1] + spacing * np.arange(1, order + 2),
    ]

    B = np.zeros((order + 1, len(t) + 1, len(x)))
    B[0, np.digitize(x, t) - 1, np.arange(len(x))] = 1

    for i in range(1, order + 1):
        for j in range(len(t) - (order + 1)):
            B[i, j] = B[i - 1, j] * (x - t[j]) / (t[j + i] - t[j]) + B[i - 1, j + 1] * (t[j + i + 1] - x) / (
                t[j + i + 1] - t[j + 1]
            )

    basis = B[-1]  # .reshape(-1, len(x))
    basis = basis[basis.sum(axis=-1) > 0]

    total_weight_per_mode = basis.sum(axis=1)
    basis = basis[total_weight_per_mode > 0.1 * total_weight_per_mode.max()]

    return basis / basis.sum(axis=0)


def cross_basis(X: list, spacing: list, order: list):
    basis = np.ones((1, 1))

    for dim, x in enumerate(X):
        x_basis = bspline_basis(x, spacing[dim], order[dim])
        basis = (x_basis[:, None] * basis).reshape(-1, len(x))
        basis = basis[basis.sum(axis=-1) > 0]

    return basis


def detrend(D, order=3):
    x = np.linspace(-1, 1, D.shape[-1])
    X = np.c_[[x**i for i in range(order + 1)]]
    A = D @ X.T @ np.linalg.inv(X @ X.T).T

    return D - A @ X


def remove_slope(D):
    return D - np.linspace(D[..., 0], D[..., -1], D.shape[-1]).T



import numpy as np


def dms_to_deg(d: float = 0, m: float = 0, s: float = 0):
    return np.radians(d + m / 60 + s / 3600)


def hms_to_deg(h: float = 0, m: float = 0, s: float = 0):
    return np.radians(h * 15 + m * 15 / 60 + s * 15 / 3600)


def deg_to_signed_dms(x: float):
    sign = -1 if x < 0 else 1
    mnt, sec = divmod(abs(x) * 3600, 60)
    deg, mnt = divmod(mnt, 60)
    return int(sign), int(deg), int(mnt), sec


def deg_to_signed_hms(x: float):
    sign = -1 if x < 0 else 1
    mnt, sec = divmod(abs(x) * 3600 / 15, 60)
    deg, mnt = divmod(mnt, 60)
    return int(sign), int(deg), int(mnt), sec


def repr_dms(x: float):
    sign, d, m, s = deg_to_signed_dms(x)
    return f"{int(sign * d):>02}°{int(m):>02}’{s:.02f}”"


def repr_hms(x: float):
    sign, h, m, s = deg_to_signed_hms(x)
    return f"{int(sign * h):>02}ʰ{int(m):>02}ᵐ{s:.02f}ˢ"


def repr_lat_lon(lat, lon):
    lat_repr = repr_dms(abs(lat)) + ("N" if lat > 0 else "S")
    lon_repr = repr_dms(abs(lon)) + ("E" if lon > 0 else "W")
    return f"{lat_repr} {lon_repr}"


def repr_phi_theta(phi, theta, frame):
    phi_deg = np.degrees(phi % 360)
    theta_deg = np.degrees(theta)
    if frame == "az_el":
        return (f"az: {phi_deg:.02f}°", f"el: {theta_deg:.02f}°")
    if frame == "ra_dec":
        return (f"ra: {repr_hms(phi_deg)}", f"dec: {repr_dms(theta_deg)}")
    if frame == "galactic":
        return (f"l: {phi_deg:.02f}°", f"b: {theta_deg:.02f}°")



from __future__ import annotations

import os

import maria
from maria.instrument import Band
from maria.io import fetch
from maria.mappers import BinMapper

here, this_filename = os.path.split(__file__)


def test_map_sim():
    map_filename = fetch("maps/cluster1.fits", refresh=True)

    f090 = Band(center=90e9, width=20e9, NET_RJ=5e-5)
    f150 = Band(center=150e9, width=30e9, NET_RJ=5e-5)

    array = {"field_of_view": 0.02, "bands": [f090, f150], "primary_size": 50}

    instrument = maria.get_instrument(array=array)

    map_filename = fetch("maps/cluster1.fits")

    input_map = maria.map.load(filename=map_filename, nu=150e9, width=0.1, center=(150, 10))

    input_map.data *= 1e3

    plan = maria.get_plan(
        scan_pattern="daisy",
        scan_options={"radius": 0.025, "speed": 0.005},  # in degrees
        duration=60,  # in seconds
        sample_rate=50,  # in Hz
        scan_center=(150, 10),
        frame="ra_dec",
    )

    sim = maria.Simulation(
        instrument,
        plan=plan,
        site="llano_de_chajnantor",
        map=input_map,
        atmosphere="2d",
    )

    tod = sim.run()

    mapper = BinMapper(
        center=(150.01, 10.01),
        frame="ra_dec",
        width=0.1,
        height=0.1,
        resolution=0.001,
        degrees=True,
        tod_preprocessing={
            "window": {"name": "tukey"},
            "filter": {"f_lower": 0.08},
            "remove_modes": {"modes_to_remove": (0,)},
            "remove_spline": {"knot_spacing": 10},
        },
        map_postprocessing={
            "gaussian_filter": {"sigma": 1},
            "median_filter": {"size": 1},
        },
        tods=[tod],
    )

    output_map = mapper.run()
    output_map.plot()
